# Copyright (c) 2019-2020 SAP SE or an SAP affiliate company. All rights reserved. This file is
# licensed under the Apache Software License, v. 2 except as noted otherwise in the LICENSE file
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
import typing
import re
import urllib.parse

import reutil

from . import cluster_domain_from_kubernetes_config

from model.base import (
    BasicCredentials,
    NamedModelElement,
    ModelValidationError,
    ModelBase,
)


class ConcourseConfig(NamedModelElement):
    '''
    Not intended to be instantiated by users of this module
    '''

    def external_url(self):
        return self.raw.get('externalUrl')

    def job_mapping_cfg_name(self):
        return self.raw.get('job_mapping')

    def concourse_uam_config(self):
        return self.raw.get('concourse_uam_config')

    def concourse_uam_cfg(self): # alias for convenience
        return self.concourse_uam_config()

    def helm_chart_default_values_config(self):
        return self.raw.get('helm_chart_default_values_config')

    def helm_chart_values(self):
        return self.raw.get('helm_chart_values', None)

    def image_pull_secret(self):
        return self.raw.get('imagePullSecret')

    def tls_secret_name(self):
        return self.raw.get('tls_secret_name')

    def kubernetes_cluster_config(self):
        return self.raw.get('kubernetes_cluster_config')

    def clamav_config(self):
        return self.raw.get('clamav_config')

    def disable_github_pr_webhooks(self):
        '''
        If set to True, the rendered concourse pull-request resources don't have webhooks configured.
        This is because of problems using webhooks on our internal Github.
        '''
        return self.raw.get('disable_webhook_for_pr', False)

    def ingress_host(self, config_factory):
        cluster_domain = cluster_domain_from_kubernetes_config(
            config_factory,
            self.kubernetes_cluster_config(),
        )
        return f'concourse.{cluster_domain}'

    def ingress_config(self):
        return self.raw.get('ingress_config')

    def ingress_url(self, config_factory):
        return 'https://' + self.ingress_host(config_factory)

    def helm_chart_version(self):
        return self.raw.get('helm_chart_version')

    def github_enterprise_host(self):
        return self.raw.get('github_enterprise_host')

    def proxy(self):
        return self.raw.get('proxy')

    def deploy_storage_class(self):
        return self.raw.get('deploy_storage_class', False)

    def github_auth_client_id(self):
        if oauth_creds := self.raw.get('oauth_client_creds'):
            return oauth_creds.get('github_auth_client_id')

    def github_auth_client_secret(self):
        if oauth_creds := self.raw.get('oauth_client_creds'):
            return oauth_creds.get('github_auth_client_secret')

    def _required_attributes(self):
        return [
            'externalUrl',
            'concourse_uam_config',
            'helm_chart_default_values_config',
            'kubernetes_cluster_config',
            'job_mapping',
            'imagePullSecret',
            'tls_secret_name',
            'ingress_config',
            'helm_chart_version',
            'helm_chart_values',
            'oauth_client_creds',
        ]

    def _optional_attributes(self):
        return {
            'clamav_config',
            'concourse_version', # TODO: Remove
            'deploy_storage_class',
            'github_enterprise_host',
            'ingress_host', # TODO: Remove
            'proxy',
        }

    def validate(self):
        super().validate()


class ConcourseEndpoint(NamedModelElement):
    def base_url(self) -> str:
        return self.raw.get('base_url')


class ConcourseTeamConfig(NamedModelElement):
    def concourse_endpoint_name(self) -> str:
        return self.raw['concourse_endpoint_name']

    def service_user(self) -> typing.Optional[BasicCredentials]:
        if local_user := self.raw.get('service_user'):
            return BasicCredentials(raw_dict=local_user)

    def team_name(self) -> str:
        return self.raw.get('team_name')

    def username(self) -> str:
        if service_user := self.service_user():
            return service_user.username()

    def password(self) -> str:
        if service_user := self.service_user():
            return service_user.passwd()

    def role(self) -> str:
        return self.raw.get('role')

    def github_auth_team(self, split: bool = False) -> typing.Optional[str]:
        """
        returns the github auth team (org:team)

        @param split: if `true` return [org, team]
        """
        git_auth_team = self.raw.get('git_auth_team')
        if split and git_auth_team:
            return git_auth_team.split(':')
        return git_auth_team

    def has_basic_auth_credentials(self) -> bool:
        if self.username() or self.password():
            return True
        return False

    def _required_attributes(self):
        return [
            'concourse_endpoint_name',
            'git_auth_team',
            'role',
            'service_user',
            'team_name',
        ]


def get_team_cfg_by_name(cc_team_cfgs, name: str) -> typing.Optional[ConcourseTeamConfig]:
    for cc_team_cfg in cc_team_cfgs:
        if cc_team_cfg.team_name() == name:
            return cc_team_cfg


def get_main_team_cfg(cc_team_cfgs):
    return get_team_cfg_by_name(cc_team_cfgs, 'main')


class ConcourseUAM(NamedModelElement):
    def local_user(self) -> typing.Optional[BasicCredentials]:
        if local_user := self.raw.get('local_user'):
            return BasicCredentials(raw_dict=local_user)

    def team_name(self) -> str:
        return self.raw.get('team_name')

    def username(self) -> str:
        if local_user := self.local_user():
            return local_user.username()

    def password(self) -> str:
        if local_user := self.local_user():
            return local_user.passwd()

    def role(self) -> str:
        return self.raw.get('role')

    def github_auth_team(self, split: bool=False) -> typing.Optional[str]:
        '''
        returns the github auth team (org:team)

        @param split: if `true` return [org, team]
        '''
        git_auth_team = self.raw.get('git_auth_team')
        if split and git_auth_team:
            return git_auth_team.split(':')
        return git_auth_team

    def has_basic_auth_credentials(self):
        if self.username() or self.password():
            return True
        return False

    def _required_attributes(self):
        return [
            'local_user',
            'role',
            'git_auth_team',
            'team_name',
        ]


class ConcourseUAMSet(NamedModelElement):
    def concourse_uams(self) -> typing.List[ConcourseUAM]:
        return [ConcourseUAM(name=name, raw_dict=raw) for name, raw in self.raw.items()]

    def concourse_uam(self, uam_name):
        for uam in self.concourse_uams():
            if uam.name() == uam_name:
                return uam
        raise ValueError(
            f"Unknown uam '{uam}'; known uams: {', '.join(self.raw.keys())}"
        )

    def main_team_uam(self):
        return self.concourse_uam('main')


class ConcourseTeam(NamedModelElement):
    def teamname(self):
        return self.name()

    def username(self):
        return self.raw.get('username')

    def password(self):
        return self.raw.get('password')

    def role(self):
        return self.raw.get('role')

    def github_auth_team(self, split: bool=False):
        '''
        returns the github auth team (org:team)

        @param split: if `true` return [org, team]
        '''
        if split and self.raw.get('git_auth_team'):
            return self.raw.get('git_auth_team').split(':')
        return self.raw.get('git_auth_team')

    def github_auth_client_id(self):
        return self.raw.get('github_auth_client_id')

    def github_auth_client_secret(self):
        return self.raw.get('github_auth_client_secret')

    def has_basic_auth_credentials(self):
        if self.raw.get('username') or self.raw.get('password'):
            return True
        return False

    def has_github_oauth_credentials(self):
        if (
            self.raw.get('git_auth_team') or
            self.raw.get('github_auth_client_id') or
            self.raw.get('github_auth_client_secret')
        ):
            return True
        return False

    def _required_attributes(self):
        yield from super()._required_attributes()
        if self.has_basic_auth_credentials():
            yield from ['username', 'password']
        if self.has_github_oauth_credentials():
            yield from ['git_auth_team', 'github_auth_client_id', 'github_auth_client_secret']

    def validate(self):
        super().validate()
        if self.has_github_oauth_credentials():
            github_org_and_team = self.github_auth_team(split=True)
            # explicitly check for expected structure, raise error if not found
            if github_org_and_team and len(github_org_and_team) == 2:
                github_org, github_team = github_org_and_team
                if github_org and github_team:
                    return
            raise ModelValidationError(
                'Invalid github-oauth team. Expected <org>/<team>, got {t}'.format(
                    t=github_org_and_team
                )
            )


class ConcourseUAMConfig(NamedModelElement):
    def teams(self) -> typing.List[ConcourseTeam]:
        return [
            ConcourseTeam(name=name, raw_dict=raw)
            for name, raw in self.raw.get('teams').items()
        ]

    def team(self, team_name: str):
        for team in self.teams():
            if team.teamname() == team_name:
                return team
        raise ValueError(
            f"Unknown team '{team_name}'; known teams: {', '.join(self.raw.get('teams').keys())}"
        )

    def main_team(self):
        return self.team('main')

    def _required_attributes(self):
        yield from super()._required_attributes()
        yield from [
            'teams',
        ]

    def validate(self):
        # We check for the existence of the 'main'-team as it is the only team that is *required* to
        # exist for any concourse server.
        if not self.main_team():
            raise ModelValidationError("No team 'main' defined.")
        # explicitly validate main team
        self.main_team().validate()


class PipelineCleanupPolicy(Enum):
    CLEANUP_EXTRA_PIPELINES = 'cleanup_extra_pipelines'
    NO_CLEANUP = 'no_cleanup'


class JobMappingSet(NamedModelElement):
    def job_mappings(self):
        return {name: JobMapping(name=name, raw_dict=raw) for name, raw in self.raw.items()}

    def job_mapping_for_repo_url(self, repo_url: str):
        for _, job_mapping in self.job_mappings().items():
            if job_mapping.matches_repo_url(repo_url):
                return job_mapping

        raise ValueError(f'no matching job mapping for {repo_url=}')

    def job_mapping_for_team_name(self, team_name: str):
        for _, job_mapping in self.job_mappings().items():
            if job_mapping.concourse_target_team() == team_name:
                return job_mapping

        raise ValueError(f'no matching job mapping for {team_name=}')

    def __iter__(self):
        return self.job_mappings().values().__iter__()

    def __getitem__(self, name: str):
        return JobMapping(name=name, raw_dict=self.raw.__getitem__(name))

    def validate(self):
        tgt_team_names = {jm.team_name() for jm in self}
        for team_name in (jm.team_name() for jm in self):
            if not team_name in tgt_team_names:
                raise ModelValidationError(f'{team_name=} must only be specified once')
            tgt_team_names.remove(team_name)


class SecretsRepo(ModelBase):
    def github_cfg(self):
        return self.raw.get('github_cfg')

    def org(self):
        return self.raw.get('org')

    def repo(self):
        return self.raw.get('repo')


def cfg_name_from_team(team_name):
    team_name = re.sub('(.)([A-Z][a-z]+)', r'\1-\2', team_name)
    team_name = re.sub('([a-z0-9])([A-Z])', r'\1-\2', team_name).lower()
    return f'cc-{team_name}-config'


def secret_cfg_name_for_team(team_name):
    return f'{team_name}_cfg'


class JobMapping(NamedModelElement):
    def team_name(self) -> str:
        # todo: use `name` attr for that (thus enforce unique mappings)
        return self.raw.get('concourse_target_team')

    def cleanup_policy(self) -> PipelineCleanupPolicy:
        return PipelineCleanupPolicy(
            self.raw.get('cleanup_policy', PipelineCleanupPolicy.CLEANUP_EXTRA_PIPELINES)
        )

    def replication_ctx_cfg_set(self) -> str:
        return self.raw.get('replication_ctx_cfg_set')

    def github_organisations(self):
        return [
            GithubOrganisationConfig(name, raw)
            for name, raw in self.raw.get('github_orgs', {}).items()
        ]

    def secret_cfg(self) -> str:
        return self.raw.get('secret_cfg')

    def secrets_repo(self) -> typing.Optional[SecretsRepo]:
        if secrets_repo := self.raw.get('secrets_repo'):
            return SecretsRepo(secrets_repo)

    def concourse_uam(self) -> str:
        return self.raw.get('concourse_uam')

    def concourse_team_cfg_name(self) -> str:
        return self.raw.get('concourse_team_cfg_name')

    def matches_repo_url(self, repo_url) -> bool:
        if not '://' in repo_url:
            repo_url = 'x://' + repo_url

        repo_url = urllib.parse.urlparse(repo_url)
        org, repo = repo_url.path[1:].split('/')

        for github_org_cfg in self.github_organisations():
            if not github_org_cfg.org_name() == org:
                continue

            if github_org_cfg.repository_matches(repo):
                return True

        return False

    def _required_attributes(self):
        return [
            'concourse_target_team',
            'replication_ctx_cfg_set',
        ]

    def _optional_attributes(self):
        return [
            'concourse_team_cfg_name',
            'concourse_uam',
            'secret_cfg',
            'secrets_repo',
        ]

    def _defaults_dict(self):
        # XXX find out why this is not honoured
        return {
            'cleanup_policy': PipelineCleanupPolicy.CLEANUP_EXTRA_PIPELINES,
        }

    def target_secret_name(self):
        '''
            k8s secret name used for replication
        '''
        return cfg_name_from_team(self.team_name())

    def target_secret_cfg_name(self):
        '''
            name of the config in the target k8s secret
        '''
        return secret_cfg_name_for_team(self.team_name())


class GithubOrganisationConfig(NamedModelElement):
    def github_cfg_name(self):
        return self.raw.get('github_cfg')

    def org_name(self):
        return self.raw.get('github_org')

    def include_repositories(self):
        return self.raw.get('include_repositories', ())

    def exclude_repositories(self):
        return self.raw.get('exclude_repositories', ())

    def repository_matches(self, repository_name: str):
        repo_filter = reutil.re_filter(
            include_regexes=self.include_repositories(),
            exclude_regexes=self.exclude_repositories(),
        )

        return repo_filter(repository_name)
